# Parses semantic net files
# keys: nodes, values: edges ([pred, target_node, sentence, 'fw'|'bw', weight])
# NODE:originnode
# EDGE:
#  PRED:pred
#  TARGET:target_node
#  SENTENCE:sentence
#  DIRECTION:'fw'or'bw'
#  WEIGHT:weight
# EDGE:
#  .....
# ********************************************
# NODE:originnode
# ...

from __future__ import nested_scopes
import re,string,os,sys,copy#,shelve
#sys.path.append('../tagger/')
#sys.path.append('../chunker/')
#sys.path.append('../lemmatiser/')
#import MontyTagger, MontyChunker, Lemmatiser

class OMCSNet:

#    shelve = 'semanticnet.shelve'
    filename = 'semanticnet.txt'
    parentdir = 'semanticnet'
    pathprefix = './'#os.getcwd()+'/../'+parentdir+'/'
    semantic_net = {}
    nodecount = 0 # tracks how many pieces of CS are being loaded
    edgecount = 0
    
    def __init__(self):
        self.thePOSTagger = None #MontyTagger.MontyTagger()
        self.theChunker = None#MontyChunker.MontyChunker()
        self.theLemmatiser = None#Lemmatiser.Lemmatiser()
        print "Loading Semantic Net from File"
        self.load()
        print "Done loading semantic net"

    def get(self,origin_node):
        edges = self.semantic_net.get(origin_node,[])
        return edges

    def get_flexible(self,concept):
        # distills a concept using the same "clean_arg" method as
        # that which was used to produce the semantic net
        cleaned = self.clean_arg(concept)
        edges = self.get(cleaned)
        return [cleaned,edges]

    def statistics(self):
        # what is the size of the semantic network?
        # #nodes, #edges/2
        nodescount = len(self.semantic_net)
        edgescount = 0
        for node in self.semantic_net.keys():
            edgescount += len(self.semantic_net.get(node,[]))
        branchingfactor = float(edgescount)/float(nodescount)/2

        print "STATISTICS on Semantic Network:\n  No. of nodes:",nodescount,"\n  No. of fw/bw edge pairs:",edgescount/2,"\n  Fw Edge branching factor:",branchingfactor
        return
    
    def clean_arg(self,arg):
        return arg # not supported in java
    
        # takes in a word or phrase and normalizes it
        # by lemmatisation, removal of determiners, adjective modifiers

        tagged = self.thePOSTagger.tag(arg)
        pairs = map(lambda x:[x.split('/')[0],x.split('/')[1]],tagged.split())
       
        output = ''
        do_strip_p = 1
        for i in range(len(pairs)):
            tag = pairs[i][1]
            word = pairs[i][0]
            if do_strip_p:
                if tag in ['TO','IN','DT','PRP','PRP$']:
                    # strip leading preposition/determiner, etc
                    continue
                else:
                    do_strip_p = 0
                    
            # strip interior dt,prp,etc, and lemmatise nouns
            if tag not in ['DT','PRP','PRP$','.',',']:
                if tag in ['NN','NNP','NNS','NNPS']:
                    word = self.theLemmatiser.lemmatise(word) 
                output += word + ' '
            
        output = string.strip(output)
        return output

    def load(self):
##        # if exists cached version aka shelve, use that
##        if self.shelve in os.listdir(self.pathprefix):
##            s = shelve.open(self.pathprefix+self.shelve,'r')
##            self.semantic_net = s
##            return
##        # else:
        
        # load file
        f = open(self.pathprefix+self.filename,'r')
        line = f.readline()
        node_set = []
        while line:
            if line[0:5] == '*****':
                # end nodeset
                self.parse_nodeset(node_set)
                node_set = []
            else:
                node_set.append(line)
            line = f.readline()

##        # store cached shelve
##        s = shelve.open(self.pathprefix+self.shelve,'n')
##        for key in self.semantic_net.keys():
##            s[key] = self.semantic_net[key]
##        s.close()

##        self.semantic_net = {}
##        # reload
##        self.load()
        return
        
    def parse_nodeset(self,node_set):
        # parses a nodeset into a dict
        # keys: nodes, values: edges ([pred, target_node, sentence, 'fw'|'bw', weight])
        # NODE:originnode
        # EDGE:
        #  PRED:pred
        #  TARGET:target_node
        #  SENTENCE:sentence
        #  DIRECTION:'fw'or'bw'
        #  WEIGHT:weight
        # EDGE:
        #  .....
        originnode = string.strip(string.split(node_set[0],'NODE:')[-1])
        edges = []
        edge = []
        for line in node_set[1:]:
            if line[:5] == 'EDGE:':
                if edge != []:
                    edges.append(edge)
                edge = []
            else:
                stripped = string.strip(string.split(line,': ')[-1])
                edge.append(stripped)

        if edge != []:
            edges.append(edge)

        edges = map(lambda edge:edge[:4]+[float(edge[4])],edges) # cast the weight as a float

        # store dict entry
        self.semantic_net[originnode] = edges

    def find_paths_between(self,nodename1,nodename2,level_tolerance=3,no_of_results=30,direction='both'):
        # finds all the ways in which two nodes are linked
        # returns all paths ordered shortest to longest, up to the no_of_results
        # direction: both==> paths may be fw or bw
        #            fw==> only fw edges, bw==> only bw edges
        # algorithm: expand undirectly from both origin
        # and destination nodes and find intersection paths
        
        print "FINDING PATHS BETWEEN",nodename1,"AND",nodename2
        concept1 = nodename1
        concept2 = nodename2
        c1paths = [[['origin',concept1,'This is the origin node','origin',1]]] # each path looks exactly like an edge
        c2paths = [[['destination',concept2,'This is the destination node','destination',1]]]
        levelcount = 1
        keep_searching_p = 1
        output_paths = []
        while levelcount < level_tolerance and keep_searching_p:
            # expand paths c1
            for path in filter(lambda x:len(x)==levelcount,c1paths): # find paths to expand
                expanded = self.get(path[0][1])
                for item in expanded:
                    if item[1] == concept1:
                        continue # new item cannot be starting node
                    newpath = copy.copy(path)
                    newpath.insert(0,list(item))
                    c1paths.append(newpath)
            
            # expand paths c2
            for path in filter(lambda x:len(x)==levelcount,c2paths):
                expanded = self.get(path[0][1])
                for item in expanded:
                    if item[1] == concept2: 
                        continue # new iteam cannot be starting node
                    newpath = copy.copy(path)
                    newpath.insert(0,list(item)) # preprend new node to each path
                    c2paths.append(newpath)
            
            # do any NEW nodes (at index 0) in c1paths match any nodes in c2paths
            for c1path in filter(lambda x:len(x)==levelcount+1,c1paths):
                for c2path in c2paths:
                    for i in range(len(c2path)):
                        c2pathnode = c2path[i][1]
                        if c1path[0][1] == c2pathnode: # match
                            c1path_rev = copy.copy(c1path)
                            c1path_rev.reverse() # reverse c1path and prepend to c1path
                            matchpath = c1path_rev + c2path[i+1:]
                            output_paths.append(matchpath)
                            if len(output_paths) >= no_of_results:
                                keep_searching_p = 0
                                break
                    if not keep_searching_p:
                        break
                if not keep_searching_p:
                    break

            # do the same for other nodes
            for c2path in filter(lambda x:len(x)==levelcount+1,c2paths):
                for c1path in c1paths:
                    for i in range(len(map(lambda x:x[0],c1path))):
                        c1pathnode = c1path[i][1]
                        if c2path[0][1] == c1pathnode: # match
                            c1path_rev = copy.copy(c1path)
                            c1path_rev.reverse() # reverse c1path and prepend to c1path
                            c1path_rev = c1path_rev[:1-i]
                            matchpath = c1path_rev + c2path
                            output_paths.append(matchpath)
                            if len(output_paths) >= no_of_results:
                                keep_searching_p = 0
                                break
                    if not keep_searching_p:
                        break
                if not keep_searching_p:
                    break

            # advance levelcount
            levelcount += 1

        # now rebuild each path to move forward from origin to destination
        if direction == 'fw':
            direction_var = ['fw']
        elif direction == 'bw':
            direction_var = ['bw']
        else:
            direction_var = ['fw','bw']
        cleaned_output_paths = []
        for path in output_paths:
            cleaned_path = [path[0]] # start with origin node path
            current_node = path[0][1]
            for edge in path[1:]:
                edge_node = edge[1]
                next_edge = filter(lambda x:x[1]==edge_node and x[3] in direction_var,self.get(current_node))
                cleaned_path.append(next_edge)
                current_node = edge_node # inc

            if 0 in map(lambda x:len(x),cleaned_path):
                # direction restriction failed,
                #don't add this path to cleaned
                pass
            else:
                cleaned_output_paths.append(cleaned_path)

        # filter duplicates
        cleaned_output_paths = map(lambda x:eval(x),filter(None, map(lambda i,d={}:(i,None)[d.has_key(i) or d.update({i:1}) or 0], map(lambda x:str(x),cleaned_output_paths))))

        return cleaned_output_paths


    def spreadingactivation(self,nodename,levels=2,tolerance=0.1,len_tolerance=20):
        # 
        # assumes node is already cleaned
        # returns [['node',0.0],['node',0.0]...etc]
        
        print "EXPAND BEGUN"
        cleaned = nodename
        annotations = []

        edges = self.get(cleaned)
        
        if edges == []:
            print cleaned,"is not a known node"
            return []
        
        annotations = map(lambda edge:edge[0:2]+edge[4:],edges)
        annotations1 = copy.copy(annotations)
        if levels == 2:
            for concept1 in annotations1:
                pred = concept1[0]
                concept = concept1[1]
                weight = concept1[2]
                
                for concept2 in self.get(concept):
                    triple = [concept2[0],concept2[1],concept2[4]*weight]
                    annotations.append(triple)

        # remove annotations == original node, by tolerance
        output_annotations = []
        for i in range(len(annotations)):
            if annotations[i][1] != cleaned and annotations[i][2] >= tolerance:
                output_annotations.append(copy.copy(annotations[i]))

        # merge items with the same key.. i.e.: ('dog',0.5) and ('dog',0.2)
        for i in range(len(output_annotations)):
            for j in range(len(output_annotations)):
               
                if i!=j and output_annotations[i][1] == output_annotations[j][1]:
                    # set the jth annotation to ()
                    # set the ith annotation to ( max(i,j) + (1-max(i,j))*min(i,j))
                    themax = max(output_annotations[i][2],output_annotations[j][2])
                    themin = min(output_annotations[i][2],output_annotations[j][2])
                    new_weight = themax + (1-themax)*themin
                    new_label = output_annotations[i][1]
                    new_pred = output_annotations[i][0]
                    output_annotations[i] = [new_pred,new_label,new_weight]
                    # set jth annotation to ()
                    output_annotations[j] = ('','',0.0)

        # remove all empty () annotations
        output_annotations = filter(lambda x:x!=('','',0.0),output_annotations)

        # remove preds
        output_annotations = map(lambda x:x[1:],output_annotations)
        
        # sort annotations by weight, descending
        output_annotations.sort(lambda x,y:int((y[1]-x[1])*1000)) # *1000 b/c sort takes int

        # enforce length tolerance
        if len(output_annotations) > len_tolerance:
            del output_annotations[len_tolerance:]

        return output_annotations

        
if __name__ == "__main__":
    print sys.argv[-1]
    s = OMCSNet()
    s.statistics()
    res = s.find_paths_between(s.clean_arg('human'),s.clean_arg('someone'),direction='both')
    print res
    res = s.spreadingactivation(s.clean_arg('dog'))
    print res
        
