# Parses semantic net files
# keys: nodes, values: edges ([pred, target_node, sentence, 'fw'|'bw', weight])
# NODE:originnode
# EDGE:
#  PRED:pred
#  TARGET:target_node
#  SENTENCE:sentence
#  DIRECTION:'fw'or'bw'
#  WEIGHT:weight
# EDGE:
#  .....
# ********************************************
# NODE:originnode
# ...

import re,string,os

class OMCSNet:

    filename = 'semanticnet.txt'
    parentdir = 'semanticnet'
    pathprefix = os.getcwd()+'/../'+parentdir+'/'
    semantic_net = {}
    
    def __init__(self):
        print "Loading Semantic Net from File"
        self.load()
        print "Done loading semantic net"

    def get(self,origin_node):
        edges = self.semantic_net.get(origin_node,[])
        return edges

    def get_flexible(self,concept):
        # distills a concept using the same "clean_arg" method as
        # that which was used to produce the semantic net
        
    def load(self):
        # load file
        f = open(self.pathprefix+self.filename,'r')
        line = f.readline()
        node_set = []
        while line:
            if line[0:5] == '*****':
                # end nodeset
                self.parse_nodeset(node_set)
                node_set = []
            else:
                node_set.append(line)
            line = f.readline()

    def parse_nodeset(self,node_set):
        # parses a nodeset into a dict
        # keys: nodes, values: edges ([pred, target_node, sentence, 'fw'|'bw', weight])
        # NODE:originnode
        # EDGE:
        #  PRED:pred
        #  TARGET:target_node
        #  SENTENCE:sentence
        #  DIRECTION:'fw'or'bw'
        #  WEIGHT:weight
        # EDGE:
        #  .....
        originnode = string.strip(string.split(node_set[0],'NODE:')[-1])
        edges = []
        edge = []
        for line in node_set[1:]:
            if line[:5] == 'EDGE:':
                if edge != []:
                    edges.append(edge)
                edge = []
            else:
                stripped = string.strip(string.split(line,': ')[-1])
                edge.append(stripped)
        # store dict entry
        self.semantic_net[originnode] = edges
        
if __name__ == "__main__":
    s = OMCSNet()
    print s.semantic_net
        
